#include <windows.h>
#include <commctrl.h>
#include <shobjidl.h>
#include <stdio.h>
#include "resource.h"

// Window size settings
int initial_size_x = 500;
int initial_size_y =  80;
int min_size_x = 245;
int min_size_y =  80;
int max_size_y =  90;

// Global variables
HINSTANCE g_hinst;
HWND g_pb;
const char g_szClassName[] = "class1";
ITaskbarList3* g_pTaskbarList = NULL;
HMENU g_menu = NULL;
HMENU g_menu_time = NULL;

enum {
    TIMER_STATUS_INACTIVE,
    TIMER_STATUS_PAUSED,
    TIMER_STATUS_RUNNING,
    TIMER_STATUS_FINISHED,
} g_timer_status = TIMER_STATUS_INACTIVE;
UINT_PTR g_timerId = 0;
int g_selected_time = ID_25MIN;
ULONGLONG g_seconds_count, g_seconds_target, g_tick_last;
BOOL g_timer_msg_shown = FALSE;
const unsigned g_timer_update_rate = 500; // Delay in ms for timer

// Updates window title bar with remaining time and status
void update_title(HWND hwnd)
{
    static char timestr[256];
    timestr[0] = '?';
    timestr[1] = '\0';
    const char* extra = "";
    switch (g_timer_status) {
        case TIMER_STATUS_INACTIVE: extra = " - Timer inactive"; break;
        case TIMER_STATUS_PAUSED:   extra = " - Paused"; break;
        case TIMER_STATUS_RUNNING:  extra = ""; break;
        case TIMER_STATUS_FINISHED: extra = " - Coffee time!"; break;
    }
    unsigned seconds = (unsigned)(g_seconds_target - g_seconds_count);
    if (seconds < 60) {
        // Show only seconds
        snprintf(timestr, sizeof timestr, "%02u%s", seconds, extra);
    } else if (seconds < 3600) {
        // Show minutes and seconds
        unsigned minutes = (seconds / 60) % 60;
        seconds %= 60;
        snprintf(timestr, sizeof timestr, "%02u:%02u%s", minutes, seconds, extra);
    } else {
        // Show hours, minutes and seconds
        unsigned hours = (seconds / 3600);
        unsigned minutes = (seconds / 60) % 60;
        seconds %= 60;
        snprintf(timestr, sizeof timestr, "%02u:%02u:%02u%s", hours, minutes, seconds, extra);
    }
    SetWindowTextA(hwnd, timestr);
}

void reset_time()
{
    g_seconds_count = 0;
    g_seconds_target = 0;
    switch (g_selected_time) {
        case ID_10MIN: g_seconds_target = 10 * 60; break;
        case ID_25MIN: g_seconds_target = 25 * 60; break;
        case ID_30MIN: g_seconds_target = 30 * 60; break;
        case ID_60MIN: g_seconds_target = 60 * 60; break;
    }
}

void reset_timer(HWND hwnd)
{
    // ... -> INACTIVE
    if (g_timer_status == TIMER_STATUS_RUNNING && g_timerId) {
        KillTimer(hwnd, g_timerId);
        g_timerId = 0;
    }
    g_timer_status = TIMER_STATUS_INACTIVE;
    if (g_pTaskbarList)
        g_pTaskbarList->SetProgressState(hwnd, TBPF_NOPROGRESS);
    SendMessage(g_pb, PBM_SETRANGE, (WPARAM)0, MAKELPARAM(0, 1));
    SendMessage(g_pb, PBM_SETPOS, (WPARAM)0, MAKELPARAM(0, 0));
    EnableMenuItem(g_menu, ID_START, MF_BYCOMMAND | MF_ENABLED);
    EnableMenuItem(g_menu, ID_PAUSE, MF_BYCOMMAND | MF_GRAYED);
    EnableMenuItem(g_menu, ID_RESUME, MF_BYCOMMAND | MF_GRAYED);
    DrawMenuBar(hwnd);
    reset_time();
    update_title(hwnd);
}

void start_timer(HWND hwnd)
{
    // INACTIVE -> RUNNING
    if (g_timer_status != TIMER_STATUS_INACTIVE)
        return;
    reset_time();
    g_tick_last = GetTickCount64();
    g_timer_status = TIMER_STATUS_RUNNING;
    g_timerId = SetTimer(hwnd, 1, g_timer_update_rate, NULL);
    if (g_pTaskbarList)
        g_pTaskbarList->SetProgressState(hwnd, TBPF_NORMAL);
    SendMessage(g_pb, PBM_SETRANGE, (WPARAM)0, MAKELPARAM(0, g_seconds_target));
    SendMessage(g_pb, PBM_SETSTEP, (WPARAM)1, 0);
    SendMessage(g_pb, PBM_SETPOS, (WPARAM)0, 0);
    EnableMenuItem(g_menu, ID_START, MF_BYCOMMAND | MF_GRAYED);
    EnableMenuItem(g_menu, ID_PAUSE, MF_BYCOMMAND | MF_ENABLED);
    EnableMenuItem(g_menu, ID_RESUME, MF_BYCOMMAND | MF_GRAYED);
    DrawMenuBar(hwnd);
    update_title(hwnd);
}

void pause_timer(HWND hwnd)
{
    // RUNNING -> PAUSED
    if (g_timer_status != TIMER_STATUS_RUNNING)
        return;
    if (g_timerId) {
        KillTimer(hwnd, g_timerId);
        g_timerId = 0;
    }
    g_timer_status = TIMER_STATUS_PAUSED;
    if (g_pTaskbarList)
        g_pTaskbarList->SetProgressState(hwnd, TBPF_PAUSED);
    EnableMenuItem(g_menu, ID_START, MF_BYCOMMAND | MF_GRAYED);
    EnableMenuItem(g_menu, ID_PAUSE, MF_BYCOMMAND | MF_GRAYED);
    EnableMenuItem(g_menu, ID_RESUME, MF_BYCOMMAND | MF_ENABLED);
    DrawMenuBar(hwnd);
    update_title(hwnd);
}

void resume_timer(HWND hwnd)
{
    // PAUSED -> RUNNING
    if (g_timer_status != TIMER_STATUS_PAUSED)
        return;
    g_tick_last = GetTickCount64();
    g_timer_status = TIMER_STATUS_RUNNING;
    g_timerId = SetTimer(hwnd, 1, g_timer_update_rate, NULL);
    if (g_pTaskbarList)
        g_pTaskbarList->SetProgressState(hwnd, TBPF_NORMAL);
    EnableMenuItem(g_menu, ID_START, MF_BYCOMMAND | MF_GRAYED);
    EnableMenuItem(g_menu, ID_PAUSE, MF_BYCOMMAND | MF_ENABLED);
    EnableMenuItem(g_menu, ID_RESUME, MF_BYCOMMAND | MF_GRAYED);
    DrawMenuBar(hwnd);
    update_title(hwnd);
}

// Used to set default font
BOOL CALLBACK SetFontCallback(HWND hwnd, LPARAM font)
{
    SendMessage(hwnd, WM_SETFONT, font, TRUE);
    return TRUE;
}

// Box displaying "Time is up" and ok button (used instead of MessageBox to center it on parent window)
INT_PTR CALLBACK TimeupBoxProc(HWND hwnd, UINT wm, WPARAM wParam, LPARAM lParam)
{
    switch (wm) {
        case WM_INITDIALOG:
        {
            NONCLIENTMETRICS metrics;
            metrics.cbSize = sizeof(NONCLIENTMETRICS);
            SystemParametersInfoA(SPI_GETNONCLIENTMETRICS, sizeof(NONCLIENTMETRICS), &metrics, 0);
            HFONT font = CreateFontIndirectA(&metrics.lfMessageFont);
            EnumChildWindows(hwnd, (WNDENUMPROC)SetFontCallback, (LPARAM)font);
            
            HWND parent = GetParent(hwnd);
            if (parent == NULL)
                parent = GetDesktopWindow();
            RECT pos, ppos;
            GetWindowRect(hwnd, &pos); 
            GetWindowRect(parent, &ppos); 
            SetWindowPos(hwnd, HWND_TOP, (ppos.left + ppos.right) / 2 - (pos.right - pos.left) / 2, ppos.top + 3, 0, 0, SWP_NOSIZE);
            return TRUE;
        }
        case WM_COMMAND:
            EndDialog(hwnd, 0);
        break;
    }
    return FALSE;
}

// Called by Windows every timer tick if the timer is active
void timer_callback(HWND hwnd)
{
    if (g_timer_status == TIMER_STATUS_RUNNING) {
        if (g_seconds_count < g_seconds_target) {
            // Countdown running, advance timer
            ULONGLONG now = GetTickCount64();
            ULONGLONG progress = now - g_tick_last;
            ULONGLONG progress_seconds = 0;
            if (progress >= 1000) {
                g_tick_last = now;
                progress_seconds = progress / 1000;
            }
            if (g_seconds_count + progress_seconds > g_seconds_target) {
                g_seconds_count = g_seconds_target;
            } else {
                g_seconds_count += progress_seconds;
            }
            // Show progress
            g_pTaskbarList->SetProgressValue(hwnd, g_seconds_count, g_seconds_target);
            SendMessage(g_pb, PBM_DELTAPOS, progress_seconds + 1, 0);
            SendMessage(g_pb, PBM_DELTAPOS, -1, 0);
            update_title(hwnd);
        }
        if (g_seconds_count >= g_seconds_target) {
            // Countdown finished. Notifiy user
            g_timer_status = TIMER_STATUS_FINISHED;
            g_pTaskbarList->SetProgressState(hwnd, TBPF_PAUSED);
            update_title(hwnd);
            SendMessage(g_pb, PBM_SETRANGE, 0, 1);
            SendMessage(g_pb, PBM_SETPOS, 1, 0);
            ShowWindow(hwnd, SW_SHOWNORMAL);
            DialogBox(GetModuleHandle(NULL), MAKEINTRESOURCE(ID_TIMEUPBOX), hwnd, TimeupBoxProc);
            reset_timer(hwnd);
        }
    } else if (g_timer_status == TIMER_STATUS_FINISHED) {
        MessageBeep(MB_ICONINFORMATION);
        FLASHWINFO fi;
        fi.cbSize    = sizeof fi;
        fi.hwnd      = hwnd;
        fi.dwFlags   = FLASHW_ALL | FLASHW_TIMERNOFG | FLASHW_TRAY;
        fi.uCount    = 0;
        fi.dwTimeout = 0;
        FlashWindowEx(&fi);
    }
}

LRESULT CALLBACK WndProc(HWND hwnd, UINT wm, WPARAM wParam, LPARAM lParam)
{
    // Initialize Taskbar
    static UINT s_uTBBC = WM_NULL;
    if (s_uTBBC == WM_NULL) {
        s_uTBBC = RegisterWindowMessage("TaskbarButtonCreated");
        ChangeWindowMessageFilter(s_uTBBC, MSGFLT_ADD);
    }
    if (wm == s_uTBBC) {
        if (!g_pTaskbarList) {
            HRESULT hr = CoCreateInstance(CLSID_TaskbarList, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&g_pTaskbarList));
            if (SUCCEEDED(hr)) {
                hr = g_pTaskbarList->HrInit();
                if (FAILED(hr)) {
                    g_pTaskbarList->Release();
                    g_pTaskbarList = NULL;
                }
            }
        }
    } else switch (wm) {
        case WM_CREATE:
        {
            // Create progress bar
            RECT area;
            GetClientRect(hwnd, &area); 
            g_pb = CreateWindowEx(0, PROGRESS_CLASS, (LPTSTR)NULL, WS_CHILD | WS_VISIBLE, 3, 3,
                                  area.right - 6, area.bottom - 6, hwnd, (HMENU)0, g_hinst, NULL);
            SendMessage(g_pb, PBM_SETRANGE, 0, MAKELPARAM(0, g_seconds_target));
            SendMessage(g_pb, PBM_SETSTEP, (WPARAM)1, 0);
            SendMessage(g_pb, PBM_SETPOS, (WPARAM)0, 0);
            
            // Create menubar
            g_menu = CreateMenu();
            AppendMenu(g_menu, MF_STRING, ID_START, "&Start");
            AppendMenu(g_menu, MF_STRING, ID_PAUSE, "&Pause");
            AppendMenu(g_menu, MF_STRING, ID_RESUME, "&Resume");
            AppendMenu(g_menu, MF_STRING, ID_RESET, "St&op");
            g_menu_time = CreatePopupMenu();
            AppendMenu(g_menu_time, MF_STRING | (ID_10MIN == g_selected_time ? MF_CHECKED : MF_UNCHECKED), ID_10MIN, "&10 minutes");
            AppendMenu(g_menu_time, MF_STRING | (ID_25MIN == g_selected_time ? MF_CHECKED : MF_UNCHECKED), ID_25MIN, "&25 minutes");
            AppendMenu(g_menu_time, MF_STRING | (ID_30MIN == g_selected_time ? MF_CHECKED : MF_UNCHECKED), ID_30MIN, "&30 minutes");
            AppendMenu(g_menu_time, MF_STRING | (ID_60MIN == g_selected_time ? MF_CHECKED : MF_UNCHECKED), ID_60MIN, "&60 minutes");
            AppendMenu(g_menu, MF_STRING | MF_POPUP, (INT_PTR)g_menu_time, "&Time");
            SetMenu(hwnd, g_menu);
            
            // Timer is inactive at first
            reset_timer(hwnd);
        }
        break;
        case WM_TIMER:
        {
            timer_callback(hwnd);
        }
        break;
        case WM_KEYUP:
        {
            // Start / pause / resume the timer by pressing the space bar
            if (LOWORD(wParam) == VK_SPACE) {
                switch (g_timer_status) {
                    case TIMER_STATUS_INACTIVE: start_timer(hwnd);  break;
                    case TIMER_STATUS_PAUSED:   resume_timer(hwnd); break;
                    case TIMER_STATUS_RUNNING:  pause_timer(hwnd);  break;
                }
            }
        }
        break;
        case WM_SIZE:
        {
            RECT area;
            GetClientRect(hwnd, &area);
            SetWindowPos(g_pb, (HWND)0, 3, 3, area.right - 6, area.bottom - 6, SWP_NOZORDER);
        }
        break;
        case WM_GETMINMAXINFO:
        {
            LPMINMAXINFO mmi = (LPMINMAXINFO)lParam;
            mmi->ptMinTrackSize.x = min_size_x;
            mmi->ptMinTrackSize.y = min_size_y;
            mmi->ptMaxTrackSize.y = max_size_y;
        }
        break;
        case WM_COMMAND:
        {
            switch(LOWORD(wParam))
            {
                case ID_START:  start_timer(hwnd);  break;
                case ID_PAUSE:  pause_timer(hwnd);  break;
                case ID_RESUME: resume_timer(hwnd); break;
                case ID_RESET:  reset_timer(hwnd);  break;
                case ID_10MIN:
                {
                    CheckMenuItem(g_menu_time, g_selected_time, MF_UNCHECKED);
                    g_selected_time = ID_10MIN;
                    CheckMenuItem(g_menu_time, g_selected_time, MF_CHECKED);
                    reset_timer(hwnd);
                }
                break;
                case ID_25MIN:
                {
                    CheckMenuItem(g_menu_time, g_selected_time, MF_UNCHECKED);
                    g_selected_time = ID_25MIN;
                    CheckMenuItem(g_menu_time, g_selected_time, MF_CHECKED);
                    reset_timer(hwnd);
                }
                break;
                case ID_30MIN:
                {
                    CheckMenuItem(g_menu_time, g_selected_time, MF_UNCHECKED);
                    g_selected_time = ID_30MIN;
                    CheckMenuItem(g_menu_time, g_selected_time, MF_CHECKED);
                    reset_timer(hwnd);
                }
                break;
                case ID_60MIN:
                {
                    CheckMenuItem(g_menu_time, g_selected_time, MF_UNCHECKED);
                    g_selected_time = ID_60MIN;
                    CheckMenuItem(g_menu_time, g_selected_time, MF_CHECKED);
                    reset_timer(hwnd);
                }
                break;
                case ID_EXIT:
                {
                    PostMessage(hwnd, WM_CLOSE, 0, 0);
                }
                break;
            }
        }
        break;
        case WM_CLOSE:
            DestroyWindow(g_pb);
            DestroyWindow(hwnd);
        break;
        case ID_EXIT:
        case WM_DESTROY:
            // Destroy widgets
            if (g_menu_time) {
                DestroyMenu(g_menu_time);
                g_menu_time = 0;
            }
            if (g_menu) {
                DestroyMenu(g_menu);
                g_menu = 0;
            }
            if (g_timerId) {
                KillTimer(hwnd, g_timerId);
                g_timerId = 0;
            }
            if (g_pTaskbarList) {
                g_pTaskbarList->Release();
                g_pTaskbarList = NULL;
            }
            PostQuitMessage(0);
        break;
        default:
            return DefWindowProc(hwnd, wm, wParam, lParam);
    }
    return 0;
}

int WINAPI WinMain(HINSTANCE hInstance, HINSTANCE hPrevInstance, LPSTR lpCmdLine, int nCmdShow)
{
    g_hinst = hInstance;
    
    // Setup window class settings
    WNDCLASSEX wc;
    wc.cbSize = sizeof(WNDCLASSEX);
    wc.style = 0;
    wc.lpfnWndProc = WndProc;
    wc.cbClsExtra = 0;
    wc.cbWndExtra = 0;
    wc.hInstance = g_hinst;
    wc.hIcon = LoadIcon(hInstance, MAKEINTRESOURCE(ID_APPICON));
    wc.hCursor = LoadCursor(NULL, IDC_ARROW);
    wc.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
    wc.lpszMenuName = NULL;
    wc.lpszClassName = g_szClassName;
    wc.hIconSm = LoadIcon(hInstance, MAKEINTRESOURCE(ID_APPICON));
    if(!RegisterClassEx(&wc)) {
        MessageBox(NULL, "RegisterClassEx failed", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    
    }
    // Create main window
    HWND hwnd = CreateWindowEx(
        WS_EX_LEFT,
        g_szClassName,
        "",
        WS_OVERLAPPEDWINDOW,
        8,
        8,
        initial_size_x, initial_size_y,
        NULL, NULL,
        g_hinst, NULL);
    if(hwnd == NULL) {
        MessageBox(NULL, "CreateWindowEx failed", "Error", MB_ICONEXCLAMATION | MB_OK);
        return 0;
    }
    ShowWindow(hwnd, nCmdShow);
    UpdateWindow(hwnd);
    
    CoCreateInstance(CLSID_TaskbarList, NULL, CLSCTX_INPROC_SERVER, IID_PPV_ARGS(&g_pTaskbarList));
    
    // Start message Loop
    MSG Msg;
    while(GetMessage(&Msg, NULL, 0, 0) > 0) {
        TranslateMessage(&Msg);
        DispatchMessage(&Msg);
    }
    return Msg.wParam;
}

